<?php

namespace Drupal\translation_extractor\Drush\Commands;

use Consolidation\SiteAlias\SiteAliasInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Extension\Extension;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\translation_extractor\Service\TranslationExtractorInterface;
use Drush\Attributes as CLI;
use Drush\Commands\DrushCommands;
use Drush\Drush;
use Symfony\Component\Console\Helper\ProgressIndicator;

/**
 * Drush commands for the Translation Extractor contrib module
 */
final class TranslationExtractorDrushCommands extends DrushCommands {

  /**
   * Drupal language manager service.
   */
  protected LanguageManagerInterface $languageManager;

  /**
   * Immutable configuration object for the extractor
   */
  protected ImmutableConfig $extractorSettings;

  /**
   * Translation extractor instance
   */
  protected TranslationExtractorInterface $extractorScanner;

  /**
   * Drupal module extension list instance
   */
  protected ModuleExtensionList $moduleExtensionListService;

  /**
   * Drupal site alias interface
   */
  protected SiteAliasInterface $drushSiteAlias;

  /**
   * @inheritDoc
   */
  public function __construct(
    LanguageManagerInterface $language_manager,
    ImmutableConfig $extractorSettings,
    TranslationExtractorInterface $extractorScanner,
    ModuleExtensionList $module_extension_list_service,
  ) {
    parent::__construct();

    $this->languageManager = $language_manager;
    $this->extractorSettings = $extractorSettings;
    $this->extractorScanner = $extractorScanner;
    $this->moduleExtensionListService = $module_extension_list_service;

    // Dependency injection not possible, because it's a drush service, not a Drupal one.
    $this->drushSiteAlias = Drush::service('site.alias.manager')->getSelf();
  }

  /**
   * Extract translations from the specified module's source code, translate those strings to
   * the given language, and finally store the module PO file to the file system.
   *
   * Following a simple and full command example:
   *
   * # Simple command example:
   *
   * drush trex-export my-module de
   *
   * This command extracts all strings from the module "my-module",
   * searches for existing "de" translations in the database, and creates the
   * PO translation file to "/your-site/web/modules/custom/my-module/translations/my-module.de.po"
   *
   * ## Set different module PO folder (optional):
   *
   * You can use the "--module-po-folder=" option to specify a different
   * folder for the generated module translation file. The filename is automatically generated based
   * on the provided module and language parameters. (e.g. "my-module.de.po")
   *
   * ## Set a suffix for the module PO filename (optional)
   *
   * Use the "--module-po-suffix=" option to add a suffix to the automatically generated filename,
   * e.g."module.de.po" becomes "module_suffix.de.po".
   *
   * ## Initially import base PO file (optional):
   *
   * Use the "--base-po-folder=" option to initially import the base
   * translation file for the language code from the given folder. If you set "de" as language code, it will
   * load the "de.po" from this folder. This will ensure an up-to-date translation
   * database in your Drupal site before extracting, translating, and generating the new module translation file.
   * If not set, the base PO file is not imported.
   *
   * ## Show module po content
   *
   * Use the "--show" flag to output the generated PO file to the console.
   *
   * # Full command example:
   *
   * drush trex-export my-module de \
   * --base-po-folder=/my-site/config \
   * --module-po-folder=/my-site/web/modules/custom/my-module/translations
   *
   * The output should be:
   *
   * Start exporting "de" translations for module "my-module"
   * Base PO file "/my-site/config/de.po" imported into database.
   * Module PO file "/my-site/web/modules/custom/my-module/translations/my-module.de.po" exported to file system.
   */
  #[CLI\Command(
    name: 'translation-extractor:export',
    aliases: ['trex-export']
  )]
  #[CLI\Argument(
    name: 'module',
    description: 'Name of a Drupal module to scan. Module must be enabled.'
  )]
  #[CLI\Argument(
    name: 'language',
    description: 'Language code, e.g. de'
  )]
  #[CLI\Option(
    name: 'module-po-folder',
    description: 'Example: "--module-po-folder=/my-site/config". Folder for the module PO file to be saved. By default, the folder is set to the translation sub folder of the module.'
  )]
  #[CLI\Option(
    name: 'module-po-suffix',
    description: 'Example: "--module-po-suffix="my-suffix" Adds the suffix to the automatically generated filename, e.g. "my-module.my-suffix.de.po".'
  )]
  #[CLI\Option(
    name: 'base-po-folder',
    description: 'Example: "--base-po-folder=/my-site/config". Folder which contains the base PO file. This file is imported to the database initially. If not set, the base PO file is not imported.'
  )]
  #[CLI\Option(
    name: 'show',
    description: 'Example: "--show". Additionally output the extracted and translated module translations to the command line.',
  )]
  #[CLI\Usage(
    name: 'trex-export my-module de',
    description: 'Extracts translation placeholders from the given module\'s source code and creates a module PO file with translations for the given language.'
  )]
  public function export(
    $module,
    $language,
    $options = [
      'module-po-folder' => NULL,
      'module-po-suffix' => NULL,
      'base-po-folder' => NULL,
      'show' => NULL,
    ],
  ): void {
    // Process general command input
    $commandSettings = $this->processGeneralCommandInput($module, $language, $options);
    if (!$commandSettings) {
      return;
    }

    [$moduleObject, $languageObject, $importBaseTranslationFile, $destinationFile] = $commandSettings;

    // Start processing - All input is validated and prepared at this point
    $this->consoleOutput('');
    $this->consoleOutput(
      'Start exporting ":language" translations for module ":module"', [
        ':module' => $module,
        ':language' => $language,
      ]
    );

    // First import base translation, if set.
    if ($importBaseTranslationFile) {
      $commandSuccessful = $this->executeDrushCommandWithSpinner(
        'locale-import',
        [
          $language,
          '--override=all',
          $options['base-po-folder'],
        ],
        dt('Importing base PO ":file" into database.', [':file' => $importBaseTranslationFile]),
        dt('Base PO file ":file" imported into database.', [':file' => $importBaseTranslationFile])
      );

      // Drush command above already output an error, just return here!
      if(!$commandSuccessful) {
        return;
      }
    }

    // Start scan and get template
    $this->extractorScanner->scan($this->extractorSettings, $module);
    $template = $this->extractorScanner->getTranslationTemplate($languageObject, []);

    // Store file
    if (file_put_contents($destinationFile, $template) === FALSE) {
      $this->consoleOutputError('File :file not writable.', [':file' => $destinationFile]);

      return;
    }

    // Done, status output
    $this->consoleOutput(
      ' ✔ Module PO file ":file" exported to file system. (:lines lines)', [
        ':file' => $destinationFile,
        ':lines' => trim($template)
          ? substr_count(trim($template), "\n") + 1
          : 0,
      ]
    );

    // All finished
    $this->consoleOutput('Done');
    $this->consoleOutput('');

    // Done, status output line 2/2
    $this->consoleOutput(
      'Use ":command" to import updated translations from ":file" back to the database.',
      [
        ':command' => $this->getExportCommandSuggestion($moduleObject, $languageObject, $options),
        ':file' => basename($destinationFile)
      ]
    );
    $this->consoleOutput('');

    // If --show is present, output the content of the module PO file to the console
    if ($options['show']) {
      $this->consoleOutput(
        'Content of ":filename":', [':filename' => basename($destinationFile)]
      );

      $this->consoleOutput($template);
    }
  }

  /**
   * Import all translation from the module PO file to the database. Optionally, the base PO file can be exported
   * to the filesystem after the module was imported using the "--base-po-folder" option.
   *
   * Following a simple and full command example:
   *
   * # Simple command example:
   *
   * drush trex-import my-module de
   *
   * This will import translations from your module PO file
   * "./path-to/modules/custom/my-module/translations/my-module.de.po" into the database.
   *
   * ## Set different module PO folder (optional):
   *
   * You can use the "--module-po-folder=" option to specify a different
   * folder to load the module PO file. The filename is automatically generated based
   * on the provided module and language parameters. (e.g. "my-module.de.po")
   *
   * ## Set a suffix for the module PO filename (optional)
   *
   * Use the "--module-po-suffix=" option to add a suffix to the automatically generated filename,
   * e.g."module.de.po" becomes "module_suffix.de.po".
   *
   * ## Export base PO file (optional):
   *
   * Use the "--base-po-folder=" to set the folder which contains the base PO file.
   * This file is exported to the filesystem after importing the module PO file.
   * If not set, the base PO file is not exported.
   *
   * # Full command example:
   *
   * drush trex-import my-module de \
   * --module-po-folder=/my-site/web/modules/custom/my-module/translations \
   * --base-po-folder=/my-site/config
   *
   * The output should be:
   *
   * Start importing "de" translations for module "my-module"
   * Module PO file "/my-site/web/modules/custom/my-module/translations/my-module.de.po" imported into database.
   * Base PO file "/my-site/config/de.po" exported to file system. (Contains all translations from database)
   */
  #[CLI\Command(
    name: 'translation-extractor:import',
    aliases: ['trex-import']
  )]
  #[CLI\Argument(
    name: 'module',
    description: 'Name of a Drupal module to scan. Module must be enabled.'
  )]
  #[CLI\Argument(
    name: 'language',
    description: 'Language code, e.g. de'
  )]
  #[CLI\Option(
    name: 'module-po-folder',
    description: 'Example: "--module-po-folder=/my-site/config". Folder of the module PO file to be imported. By default, the folder is set to the translation sub folder of the module.'
  )]
  #[CLI\Option(
    name: 'module-po-suffix',
    description: 'Example: "--module-po-suffix="my-suffix" Adds the suffix to the automatically generated filename, e.g. "my-module.my-suffix.de.po".'
  )]
  #[CLI\Option(
    name: 'base-po-folder',
    description: 'Example: "--base-po-folder=/my-site/config". Folder which contains the base PO file. This file is exported to the filesystem after importing the module PO file. If not set, the base PO file is not exported.'
  )]
  #[CLI\Usage(
    name: 'trex-import my-module de',
    description: 'Import all translation from the module PO file.'
  )]
  public function import(
    $module,
    $language,
    $options = [
      'module-po-folder' => NULL,
      'module-po-suffix' => NULL,
      'base-po-folder' => NULL,
    ],
  ): void {
    // Process general command input
    $commandSettings = $this->processGeneralCommandInput($module, $language, $options);
    if (!$commandSettings) {
      return;
    }

    [,,$importBaseTranslationFile, $destinationFile] = $commandSettings;

    // Start processing - All input is validated and prepared at this point
    $this->consoleOutput('');
    $this->consoleOutput(
      'Start importing ":language" translations for module ":module"', [
        ':module' => $module,
        ':language' => $language,
      ]
    );

    // Import module PO file to database
    $commandSuccessful = $this->executeDrushCommandWithSpinner(
      'locale-import',
      [
        $language,
        '--override=all',
        $destinationFile,
      ],
      dt('Importing all module translation from file ":file" into database.', [':file' => $destinationFile]),
      dt('Module PO file ":file" imported into database.', [':file' => $destinationFile])
    );

    // Drush command above already output an error, just return here!
    if(!$commandSuccessful) {
      return;
    }

    // Export base PO file to filesystem
    if ($importBaseTranslationFile) {
      $commandSuccessful = $this->executeDrushCommandWithSpinner(
        'locale-export',
        [
          $language,
        ],
        dt('Exporting all translations from database into base PO ":file".', [':file' => $importBaseTranslationFile]),
        dt('Base PO file ":file" exported to file system. (Contains all translations from database)', [':file' => $importBaseTranslationFile]),
        $commandOutput
      );

      // Drush command above already output an error, just return here!
      if(!$commandSuccessful) {
        return;
      }

      // Store command output to base PO file
      file_put_contents($importBaseTranslationFile, $commandOutput);
    }

    // All finished
    $this->consoleOutput('Done');
    $this->consoleOutput('');
  }

  /**
   * Execute a Drush command and add a spinner, including a start and finished message.
   *
   * @param string $command
   *   Name of the drush command to run
   * @param array $commandParameters
   *   Command parameters
   * @param string $messageRunning
   *   Message to output while the drush command still runs
   * @param string $messageFinished
   *   Message to output after the drush command finished
   * @param string|null $commandOutput
   *   Reference, used to return the command output
   *
   * @return bool
   *   Was this drush command successful or not.
   */
  private function executeDrushCommandWithSpinner(
    string $command,
    array $commandParameters,
    string $messageRunning,
    string $messageFinished,
    null|string &$commandOutput = NULL,
  ): bool {
    $drushProcess = Drush::drush($this->drushSiteAlias, $command, $commandParameters);

    $drushProcess->start();

    $progressIndicator = new ProgressIndicator(self::output());

    $progressIndicator->start($messageRunning);

    while ($drushProcess->isRunning()) {
      $progressIndicator->advance();
    }

    $progressIndicator->finish($messageFinished);

    if (!$drushProcess->isSuccessful()) {
      $this->consoleOutputError('Import error: :error', [':error' => $drushProcess->getErrorOutput()]);

      return FALSE;
    }

    // Store output in reference
    $commandOutput = $drushProcess->getOutput();

    return TRUE;
  }

  /**
   * Check, if the "locale" module is enabled
   *
   * @return bool
   */
  private function isCoreModuleLocaleEnabled(): bool {
    // Is required local module enabled?
    if (!$this->moduleExtensionListService->exists('locale')) {
      $this->consoleOutputError('Drupal core module "locale" ("Interface Translation") required. Please enable it.');

      return FALSE;
    }

    return TRUE;
  }

  /**
   * Get the module object for a given module name
   *
   * @param $module
   *   Module name
   *
   * @return \Drupal\Core\Extension\Extension|bool
   */
  private function getModule(string $module): Extension|bool {
    // Get Drupal module object for given input module name.
    $moduleObject = $this->extractorScanner->getModule($module);

    if (!$moduleObject) {
      $this->consoleOutputError('The module ":module" is not enabled.', [':module' => $module]);

      return FALSE;
    }

    return $moduleObject;
  }

  /**
   * Get the language object for a given language code, e.g. "de".
   *
   * @param string $language
   *   Language code, e.g. "de"
   *
   * @return \Drupal\Core\Language\LanguageInterface|bool
   */
  private function getLanguage(string $language): LanguageInterface|bool {
    // Get language object for given input language code.
    $languageObject = $this->languageManager->getLanguage($language);

    if (!$languageObject) {
      $this->consoleOutputError(
        'The Language ":language" is not set for your site. (Configured languages: :languages)',
        [
          ':language' => $language,
          ':languages' => join(', ', array_keys($this->languageManager->getLanguages()))
        ]
      );

      return FALSE;
    }

    return $languageObject;
  }

  private function getBaseTranslationFilename(
    string $basePoFolder,
    LanguageInterface $languageObject,
  ): string|bool {
    // If --base-po-folder option is set, check if the base translation file is available.
    $importBaseTranslationFile = sprintf(
      '%s/%s.po',
      rtrim($basePoFolder, DIRECTORY_SEPARATOR),
      $languageObject->getId()
    );

    if (!file_exists($importBaseTranslationFile)) {
      $this->consoleOutputError('Can\'t load base translation file ":file".', [':file' => $importBaseTranslationFile]);

      return FALSE;
    }

    return $importBaseTranslationFile;
  }

  /**
   * Get the filename for the module translations. If $modulePoFolder is not set, it generates
   * the filename using the default module folder (+ translations sub folder).
   *
   * If $modulePoFolder is set, it uses this folder as base and just adds the filename of the po file.
   *
   * @param string|null $modulePoFolder
   * @param \Drupal\Core\Extension\Extension $moduleObject
   * @param \Drupal\Core\Language\LanguageInterface $languageObject
   * @param string|null $modulePoSuffix
   *
   * @return string|bool
   */
  private function getModuleTranslationFilename(
    string|null $modulePoFolder,
    Extension $moduleObject,
    LanguageInterface $languageObject,
    string|null $modulePoSuffix
  ): string|bool {
    // Check input option "folder". If not set, generate folder for po based on module name.
    if ($modulePoFolder !== NULL) {
      if (!is_dir($modulePoFolder)) {
        $this->consoleOutputError('The export folder ":folder" does not exist.', [':folder' => $modulePoFolder]);

        return FALSE;
      }

      $folder = rtrim($modulePoFolder, DIRECTORY_SEPARATOR);
    }
    else {
      // Store the po file in the module's translations folder.
      $folder = sprintf('%s/%s/translations', DRUPAL_ROOT, $moduleObject->getPath());

      if (!is_dir($folder)) {
        mkdir($folder);
      }
    }

    $moduleName = $moduleObject->getName();

    // Add the suffix, if set (e.g. my-module.de.po becomes module_prefix.de.po)
    $moduleName .= ($modulePoSuffix)
      ? sprintf('_%s', $modulePoSuffix)
      : '';

    // Generate final filename
    return sprintf('%s/%s.%s.po', $folder, $moduleName, $languageObject->getId());
  }

  /**
   * Generate a drush command suggestion
   *
   * @param \Drupal\Core\Extension\Extension $moduleObject
   * @param \Drupal\Core\Language\LanguageInterface $languageObject
   * @param array $options
   *
   * @return string
   */
  private function getExportCommandSuggestion(
    Extension $moduleObject,
    LanguageInterface $languageObject,
    array $options
  ): string {
    $commandSuggestion = dt(
      'drush trex-import :module :language',
      [
        ':module' => $moduleObject->getName(),
        ':language' => $languageObject->getId(),
      ]
    );

    if ($options['module-po-folder']) {
      $commandSuggestion .= dt(' --module-po-folder=:folder', [':folder' => $options['module-po-folder']]);
    }

    if ($options['module-po-suffix']) {
      $commandSuggestion .= dt(' --module-po-suffix=:suffix', [':suffix' => $options['module-po-suffix']]);
    }

    if ($options['base-po-folder']) {
      $commandSuggestion .= dt(' --base-po-folder=:folder', [':folder' => $options['base-po-folder']]);
    }

    return $commandSuggestion;
  }

  /**
   * Process the command input shared between export and import command
   *
   * @param string $module
   *   Module name
   * @param string $language
   *   Language code, e.g. "de"
   * @param array $options
   *   Command options, e.g. "module-po-folder"
   *
   * @return array|bool
   *   Array contains [$moduleObject, $languageObject, $importBaseTranslationFile, $destinationFile]
   */
  private function processGeneralCommandInput(
    string $module,
    string $language,
    array $options,
  ): array|bool {
    // Check, if the required Drupal core "locale" module is activated
    if (!$this->isCoreModuleLocaleEnabled()) {
      return FALSE;
    }

    // Get the module object
    $moduleObject = $this->getModule($module);
    if (!$moduleObject) {
      return FALSE;
    }

    // Get the language object
    $languageObject = $this->getLanguage($language);
    if (!$languageObject) {
      return FALSE;
    }

    // Generate the base translation PO filename
    $importBaseTranslationFile = NULL;
    if ($options['base-po-folder']) {
      $importBaseTranslationFile = $this->getBaseTranslationFilename($options['base-po-folder'], $languageObject);
      if (!$importBaseTranslationFile) {
        return FALSE;
      }
    }

    // Generate the module translation PO filename
    $destinationFile = $this->getModuleTranslationFilename(
      $options['module-po-folder'],
      $moduleObject,
      $languageObject,
      $options['module-po-suffix']
    );
    if (!$destinationFile) {
      return FALSE;
    }

    return [
      $moduleObject,
      $languageObject,
      $importBaseTranslationFile,
      $destinationFile,
    ];
  }

  /**
   * Helper for console log
   *
   * @param string $message
   *  Message text with placeholders like :content
   * @param array $placeholderReplacements
   *  Array with placeholder replacements [':content' => 'The content']
   *
   * @return void
   */
  private function consoleOutput(
    string $message,
    array $placeholderReplacements = [],
  ): void {
    $this->output()->writeln(dt($message, $placeholderReplacements));
  }

  /**
   * Helper for console error log
   *
   * @param string $message
   *  Message text with placeholders like :content
   * @param array $placeholderReplacements
   *  Array with placeholder replacements [':content' => 'The content']
   *
   * @return void
   */
  private function consoleOutputError(
    string $message,
    array $placeholderReplacements = [],
  ): void {
    $this->logger()->error(dt($message, $placeholderReplacements));
  }

}
